/* $Id: sample_load.c,v 3.6 2003/01/27 18:45:50 folkert Exp folkert $

    sample_load, a load-logger
    Copyright (C) 2000  Folkert van Heusden (folkert@vanheusden.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Log: sample_load.c,v $
Revision 3.6  2003/01/27 18:45:50  folkert
bug-fixes


*/

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/vfs.h>
#if 0
	#include <linux/proc_fs.h>		/* for PROC_SUPER_MAGIC */
#else
	#define PROC_SUPER_MAGIC 0x9fa0
#endif

typedef struct
{
	pid_t	pid;
	char	*cmd;
	ulong	ticks_used;
	float	percentage;	/* ugly; should've been in an extra table [pid, percentage] */
	struct	timeval timestamp; /* time of sample */
}
procent;

typedef struct
{
	procent	*pl;
	int	nEntries;
	int	nLength;
}
proclist;

double *dataL = NULL;	/* load */
time_t *dataT = NULL;	/* time of sample */
char  **dataP = NULL;	/* process-names list */
pid_t  *dataPID = NULL;	/* pid-list */
float *dataPERC = NULL;	/* percentage list */
long *dataM_p = NULL;	/* memory physical */
long *dataM_s = NULL;	/* memory swap */
int N = 0;
char logprocess = 0;
char logmem = 0;
char logtosql = 0;
char shortmemamount = 0;
char *dt_format = NULL;	/* date/time format */
double exit_load = 0.0;
long kill_mem = -1;	/* -1 is; do not kill anyone (...when using too much memory) */
char do_kill_cpu = 0;
float kill_cpu = 999;	/* kill anyone (...when consuming too much cpu-time) */
proclist *prevproclist = NULL;
proclist *curproclist = NULL;
int interval = 60;

char *filename = NULL;

void delete_proclist(proclist *list)
{
	int loop;

	if (!list) return;

	for(loop=0; loop<list -> nEntries; loop++)
	{
		free((list -> pl)[loop].cmd);
	}

	free(list -> pl);

	free(list);
}

int add_to_proclist(proclist **list, pid_t pid, char *cmd, ulong ticks, struct timeval timestamp)
{
	/* create list */
	if (!*list)
	{
		*list = (proclist *)malloc(sizeof(proclist));
		if (!*list)
		{
			syslog(LOG_WARNING, "Memory allocation problem [create proclist]");
			return -1;
		}

		(*list) -> nEntries = 0;
		(*list) -> nLength  = 0;
		(*list) -> pl       = NULL;
	}

	/* grow list */
	if ((*list) -> nEntries == (*list) -> nLength)
	{
		procent *dummy;
		int newLength = ((*list) -> nLength << 1) + 2;

		dummy = realloc((*list) -> pl, newLength * sizeof(procent));

		if (!dummy)
		{
			syslog(LOG_WARNING, "Memory allocation problem [resize proclist]");
			return -1;
		}

		(*list) -> pl = dummy;
		(*list) -> nLength = newLength;
	}

	/* add entry to list */
	((*list) -> pl)[(*list) -> nEntries].pid = pid;
	((*list) -> pl)[(*list) -> nEntries].cmd = strdup(cmd);
	if (((*list) -> pl)[(*list) -> nEntries].cmd == NULL)
	{
		syslog(LOG_WARNING, "Memory allocation problem! [add to proclist]");
		return -1;
	}
	((*list) -> pl)[(*list) -> nEntries].ticks_used = ticks;
	((*list) -> pl)[(*list) -> nEntries].percentage = 0.0;
	((*list) -> pl)[(*list) -> nEntries].timestamp = timestamp;

	/* update list administration */
	(*list) -> nEntries++;

	return 0;
}

long getfield(char *string, int nr, char **newp)
{
	int loop;

	for(loop=0; loop<nr; loop++)
	{
		string = strchr(string, ' ');
		if (!string)
		{
			return -1;
		}

		while (*string == ' ') string++;
	}

	/* do we like to know the end of the string? */
	if (newp)
	{
		*newp = string;

		/* find end of string */
		while (*(*newp) != ' ' && *(*newp) != 0x00) (*newp)++;

		/* find start of new string */
		while (*(*newp) == ' ') (*newp)++;
	}

	return atoi(string);
}

int findProcInList(proclist *list, int prev_index, pid_t pidtolookfor, char *cmd)
{
	int loop;

	for(loop=0; loop<list -> nEntries; loop++)
	{
		int cur_index = (prev_index + loop) % (list -> nEntries);

		if ((list -> pl)[cur_index].pid == pidtolookfor &&
		strcmp((list -> pl)[cur_index].cmd, cmd) == 0)
		{
			return cur_index;
		}
	}

	return -1;
}

int kill_process(pid_t pid, char *proc, char *reason)
{
	if (pid == 1)
	{
		syslog(LOG_WARNING, "Were about to kill `init'! -> using too much %s", reason);
	}
	else
	{
		/* send KILL signal */
		if (kill(pid, SIGKILL) == -1)
		{
			syslog(LOG_WARNING, "Could not kill process %d (%s)! (using too much %s) [%d]", pid, proc, reason, errno);

			return -1;
		}
		else
		{
			syslog(LOG_WARNING, "Process %d (%s) was using too much %s. Process is succesfully terminated.", pid, proc, reason);
		}
	}

	return 0;
}

int calc_cputime_percentages(proclist *cur, proclist *prev, int *index_biggest)
{
	int loop;
	float max_perc = -1.0;

	*index_biggest = -1;

	/* no cur-list? can't do a thing */
	if (!cur || !prev)
	{
		return -1;
	}

	for(loop=0; loop<cur -> nEntries; loop++)
	{
		int prev_index = findProcInList(prev, loop, (cur -> pl)[loop].pid, (cur -> pl)[loop].cmd);
		float percentage = -1.0;

		/* calculate cpu time usage in % */
		if (prev_index != -1)	/* don't do new processes; they need to do some init and all */
		{
			double now, then;
			float	total_ticks;

			now = (double)(cur -> pl)[loop].timestamp.tv_sec
				+ ((double)(cur -> pl)[loop].timestamp.tv_usec * 1e-6);
			then = (double)(prev -> pl)[prev_index].timestamp.tv_sec
				+ ((double)(prev -> pl)[prev_index].timestamp.tv_usec * 1e-6);

			total_ticks = (now - then) * (double)HZ;

			percentage = ((float)( 
				(cur -> pl)[loop].ticks_used - (prev -> pl)[prev_index].ticks_used 
							) / total_ticks) * (float)100.0;
		}
		(cur -> pl)[loop].percentage = percentage;

		/* big-time user? */
		if (percentage > max_perc)
		{
			max_perc = percentage;
			*index_biggest = loop;
		}

		/* see if it's eating up too much */
		if (percentage >= kill_cpu && do_kill_cpu)
		{
			/* kill if not pid == 1 */
			if (kill_process((cur -> pl)[loop].pid, (cur -> pl)[loop].cmd, "processortime") == -1)
			{
				return -1;
			}
		}
	}

	return 0;
}

int scanProc(void)
{
	DIR *dirh;
	struct dirent *direntry;
	proclist *new = NULL;

	/* let's not do all those things when we don't care about it anyway */
	if (kill_mem == -1 && do_kill_cpu && logprocess == 0)
	{
		return 0;
	}

	/* open dir */
	dirh = opendir("/proc");
	if (!dirh)
	{
		syslog(LOG_ERR, "Cannot open dir /proc. [%d]", errno);
		return -1;
	}

	/* delete 'previous list' */
	delete_proclist(prevproclist);

	/* scan dir */
	while ((direntry = readdir(dirh)))
	{
		/* general variables */
		pid_t	pid;
		char	fname_buffer[1024];
		FILE	*stat_file;
		struct	timeval timestamp;
		char	line_buf[4096], *dummy;
		int	nread = 0;
		/* process-related variables */
		int	s_pid;
		char	*s_comm;
		char	s_state;
		/* ... */
		char	s_flags;
		/* ... */
		int	s_minflt;
		int	s_cminflt;
		int	s_majflt;
		int	s_cmajflt;
		int	s_utime;
		int	s_stime;
		int	s_cutime;
		int	s_cstime;
		int	s_counter;
		/* ... */
		int	s_vsize;
		int	s_rss;
		int	s_rlim;

		/* digit? */
		if (!isdigit(direntry -> d_name[0]))
			continue;

		/* convert filename to PID */
		pid = atoi(direntry -> d_name);

		/* 0? */
		if (pid == 0)
			continue;

		/* me? */
		if (pid == getpid())
			continue;

		/* existing? */
		if (kill(pid, 0) == -1)
		{
			if (errno == ESRCH)
				continue;
		}

		/* try to open */
		snprintf(fname_buffer, 1024, "/proc/%s/stat", direntry -> d_name);
		fname_buffer[1024-1] = 0x00;
		stat_file = fopen(fname_buffer, "r");
		/* could not open; maybe context-switched happend & process has gone away then */
		if (!stat_file)
			continue;

		/* read data */
		nread = fread(line_buf, 1, 4095, stat_file);
		if (nread == 0)
			goto dirty_hack;
		line_buf[nread] = 0x00;

		/* now get timestamp a.s.a.p. */
		if (gettimeofday(&timestamp, NULL) == -1)
			goto dirty_hack;	/* yeah, well, that failed :-/ */

		/* verify the PID */
		s_pid = atoi(line_buf);
		if (s_pid != pid)
			goto dirty_hack;

		/* get commandline from string */
		s_comm = strchr(line_buf, '(');
		if (!s_comm)
			goto dirty_hack;
		s_comm++; /* remove '(' */
		/* find end of commandline */
		dummy = strchr(line_buf, ')');
		if (!dummy)
			goto dirty_hack;
		*dummy = 0x00;

		/* dummy points to end of commandline; search for beginning of first token */
		dummy++;

		/* get process state */
		s_state = getfield(dummy, 0, &dummy);

		/* ... */

		/* get process flags */
		s_flags = getfield(dummy, 5, &dummy);

		/* */
		s_minflt = getfield(dummy, 0, &dummy);

		/* */
		s_cminflt = getfield(dummy, 0, &dummy);

		/* */
		s_majflt = getfield(dummy, 0, &dummy);

		/* */
		s_cmajflt = getfield(dummy, 0, &dummy);

		/* */
		s_utime = getfield(dummy, 0, &dummy);

		/* */
		s_stime = getfield(dummy, 0, &dummy);

		/* */
		s_cutime = getfield(dummy, 0, &dummy);

		/* */
		s_cstime = getfield(dummy, 0, &dummy);

		/* */
		s_counter = getfield(dummy, 0, &dummy);

		/* vsize */
		s_vsize = getfield(dummy, 4, &dummy);

		/* rss */
		s_rss = getfield(dummy, 0, &dummy);

		/* rlim */
		s_rlim = getfield(dummy, 0, &dummy);

		/* ... */

		/* verify memory usage (and only kill when pid != 1 (let's not bring down the system)) */
		if (((s_vsize + s_rss) /* * getpagesize() */) > kill_mem && kill_mem != -1)	/* process using too much memory? */
		{
			if (s_pid == 1)
			{
				syslog(LOG_WARNING, "Were about to kill `init'! -> using too much memory");
			}
			else
			{
				/* send KILL signal */
				if (kill(s_pid, SIGKILL) == -1)
				{
					syslog(LOG_WARNING, "Could not kill process %d (%s)! (using too much memory) [%d]", s_pid, s_comm, errno);
				}
				else
				{
					syslog(LOG_WARNING, "Process %d (%s) was using too much memory. Process is succesfully terminated.", s_pid, s_comm);
				}
			}

			/* after killing a process, there's not so much need of logging it */
			goto dirty_hack;
		}

		/* add entry to proclist (for further investigation/logging) */
		if (do_kill_cpu || logprocess != 0)
		{
			/* add to proclist */
			if (add_to_proclist(&new, s_pid, s_comm, s_utime + s_stime, timestamp) == -1)
			{
				/* failed */
				goto dirty_hack;
			}
		}

dirty_hack:
		fclose(stat_file);
	}

	prevproclist = curproclist;
	curproclist = new;

	/* close directory */
	if (closedir(dirh) == -1)
	{
		syslog(LOG_WARNING, "Problem while closing /proc! [%d]", errno);
	}

	return 0;
}

double loadAvg()
{
	double	avg = -1.0;	/* load avg of -1.0 indicates error */
	FILE	*file;
	char	line[81];
    
	file = fopen("/proc/loadavg", "r");
	if (file == NULL)
	{
		syslog(LOG_ERR, "Problem opening '/proc/loadavg'! (while retrieving current load)");
		exit(1);
	}

	if (fgets(line, 80, file))
	{
		avg = atof(line);
	}
	fclose(file);

	return avg;
}

void memInUse(long *physical, long *swap)
{
	FILE	*file;
	char	line[161];
	char	*state;
    
	file = fopen("/proc/meminfo", "r");
	if (file == NULL)
	{
		syslog(LOG_ERR, "Problem opening '/proc/meminfo'! (while retrieving current memory usage)");
		exit(1);
	}

	/* ignore first line */
	state = fgets(line, 160, file);
	if (state != NULL)
	{
		/* second value in second line is physical mem in use */
		state = fgets(line, 160, file);
	}
	if (state != NULL)
	{
		*physical = getfield(line, 2, NULL);

		/* second value in second line is physical mem in use */
		state = fgets(line, 160, file);
	}
	if (state != NULL)
	{
		*swap = getfield(line, 2, NULL);
	}

	fclose(file);
}

void writeMem(long value, FILE *stream)
{
	if (shortmemamount)
	{
		if (value > ((1024*1024*1024)-1))	/* GB */
		{
			fprintf(stream, " %ldGB", (value + ((1024*1024*1024)-1)) >> 30);
		}
		else if (value > ((1024*1024)-1))	/* MB */
		{
			fprintf(stream, " %ldMB", (value + ((1024*1024)-1)) >> 20);
		}
		else if (value > (1024-1))		/* KB */
		{
			fprintf(stream, " %ldKB", (value + (1024-1)) >> 10);
		}
		else
		{
			fprintf(stream, " %ldB", value);
		}
	}
	else
	{
		fprintf(stream, " %ld", value);
	}
}

void writedata(void)
{
	FILE *out;
	int loop;

	/* open file */
	if (strcmp(filename, "-") == 0)	/* console */
		out = fopen("/dev/tty", "w+");
	else
		out = fopen(filename, "a+");
	if (!out)
	{
		syslog(LOG_ERR, "Could not open file %s!", filename);
		exit(1);
	}

	/* write data */
	for(loop=0; loop<N; loop++)
	{
printf("%d %d\n", N, loop);

		/* log to sql-queries-file? */
		if (logtosql)
		{
			struct tm *now=localtime(&dataT[loop]);
                        fprintf(out, "INSERT INTO load_log(time_stamp, system_load, physical_mem, swap_mem, proc_id, proc_name, proc_cpu_usage_percentage) VALUES(%d%02d%02d%02d%02d%02d, %f, %ld, %ld, %d, '%s', %3.3f);\n", now->tm_year+1900,now->tm_mon+1,now->tm_mday,now->tm_hour, now->tm_min, now->tm_sec, dataL[loop], dataM_p[loop], dataM_s[loop], dataPID[loop], dataP[loop], dataPERC[loop]);

			free(dataP[loop]);
		}
		else
		{	 /* all other cases */

			if (dt_format == NULL)
			{
				/* just a timestamp */
				fprintf(out, "%ld", dataT[loop]);
			}
			else
			{
				/* formatted date/time-string */
				char buffer[256];
				struct tm *tm = localtime(&(dataT[loop]));

				(void)strftime(buffer, 255, dt_format, tm);

				fprintf(out, "%s", buffer);
			}

			fprintf(out, " %f", dataL[loop]);

			if (logmem == 1)
			{
				writeMem(dataM_p[loop], out);
				writeMem(dataM_s[loop], out);
			}
			if (logprocess == 1 && dataP[loop] != NULL)
			{
				fprintf(out, " %d %f (%s)", dataPID[loop], dataPERC[loop], dataP[loop]);
				free(dataP[loop]);
			}

			fprintf(out, "\n");
		}
	}

	fclose(out);

	N = 0;
}

void sighandler(int sig)
{
	/* ignore signals for a while */
	signal(SIGHUP, SIG_IGN);

	writedata();

	/* only flush data? continue */
	if (sig == SIGHUP)
	{
		signal(SIGHUP, sighandler);
	}
	else
	{
		exit(0);
	}
}

int main(int argc, char *argv[])
{
	int loop, maxN = 100;
	double load_log = 0.0;
	char bg = 0;
	struct statfs fss;

	fprintf(stderr, "*** sample_load v3.8 ***\n\nWritten for Nexus B.V. (www.nexusonline.nl) for the www.geldshop.nl project\nby F.J.J. van Heusden (folkert@vanheusden.com).\n\n");
	fprintf(stderr, "sample_load comes with ABSOLUTELY NO WARRANTY\n");
	fprintf(stderr, "This is free software, and you are welcome to redistribute it\n");
	fprintf(stderr, "under certain conditions; see 'license.txt' for more details.\n");
	fprintf(stderr, "The file 'license.txt' should've been included in the distribution\n");
	fprintf(stderr, "of this program. If it's not there or you couldn't find it, please\n");
	fprintf(stderr, "visit http://www.vanheusden.com/Linux/sample_load.php3 and download\n");
	fprintf(stderr, "the official release from that site.\n\n");

	/* verify number of arguments */
	if (argc < 3)
	{
		fprintf(stderr, "Usage: %s [options]\n", argv[0]);
		fprintf(stderr, "Options:\n");
		fprintf(stderr, " -n x	   Number of samples to keep in memory before writing to file\n");
		fprintf(stderr, " -d \"...\"  Date/time format-string; 'man 3 strftime' for specification of format\n");
		fprintf(stderr, " -dsql     Will generate SQL-statements for feeding to database. Also selects -u and -b");
		fprintf(stderr, "           Will also disable anything specified with -d \"...\"\n");
		fprintf(stderr, " -m x      Specifies at what load this application should terminate. Omit to disable this feature.\n");
		fprintf(stderr, " -u        Log memory- & swap-usage\n");
		fprintf(stderr, " -a        Write memory-amounts as KB/MB/GB (for '-u' option)\n");
		fprintf(stderr, " -b        Log username+pid+cmdline of the process that used the most CPU time in the last interval.\n");
		fprintf(stderr, " -km x     Kill processes using more then 'x' B/KB/MB/GB (B for byte is optional)\n");
		fprintf(stderr, " -kl x     Kill processes using more then 'x' %% of CPU time in the sample interval (use with care!)\n");
		fprintf(stderr, " -o \"...\"  File to write to (you MUST specify a filename!). Use '-' for console.\n");
		fprintf(stderr, " -i x      Sample interval.\n");
		fprintf(stderr, " -l x      Minimum load to store data.\n");
		fprintf(stderr, " -f        Run as a background-process.\n");
		fprintf(stderr, "Example:\n");
		fprintf(stderr, "%s -n 100 -m 15.0 -u -o /tmp/load.log -i 60 -l 0.0 -d \"%%m-%%d-%%Y %%H:%%M:%%S\"\n", argv[0]);
		fprintf(stderr, "kill -HUP {sample_load PID} flushes the buffered data\n");
		fprintf(stderr, "SQL-Statement for creating table in database for '-dsql'-option:\n");
		fprintf(stderr, "CREATE TABLE load_log(time_stamp timestamp not null, system_load float(3,5) not null, physical_mem int(8), swap_mem int(8), proc_id int(5), proc_name varchar(255), proc_cpu_usage_percentage float(4,3), primary key (time_stamp));\n");

		return 1;
	}

	/* we really need /proc */
	if (statfs("/proc", &fss) != 0 || fss.f_type != PROC_SUPER_MAGIC)
	{
		fprintf(stderr, "This program needs proc-fs mounted on /proc to work.\n");
		return -1;
	}

	/* process arguments */
	for(loop=1; loop<argc; loop++)
	{
		if (strcmp(argv[loop], "-n") == 0)
		{
			maxN = atoi(argv[loop+1]);
			if (maxN < 1)
			{
				fprintf(stderr, "Buffersize must be at least '1'.\n");
				return 1;
			}
			loop++;

			fprintf(stderr, "Will buffer %d samples.\n", maxN);
		}
		else if (strcmp(argv[loop], "-km") == 0)
		{
			int pos;
			/* convert B/KB/MB/GB to value -> kill_mem */
			pos = strlen(argv[loop+1] - 2);

			if (strcasecmp(&argv[loop+1][pos], "GB") == 0)
			{
				argv[loop+1][pos] = 0x00;
				kill_mem = atoi(argv[loop+1]) * 1024 * 1024 * 1024;
			}
			else if (strcasecmp(&argv[loop+1][pos], "MB") == 0)
			{
				argv[loop+1][pos] = 0x00;
				kill_mem = atoi(argv[loop+1]) * 1024 * 1024;
			}
			else if (strcasecmp(&argv[loop+1][pos], "KB") == 0)
			{
				argv[loop+1][pos] = 0x00;
				kill_mem = atoi(argv[loop+1]) * 1024;
			}
			else if (strcasecmp(&argv[loop+1][pos+1], "B") == 0)
			{
				argv[loop+1][pos] = 0x00;
				kill_mem = atoi(argv[loop+1]);
			}
			else
			{
				kill_mem = atoi(argv[loop+1]);
			}

			loop++;
			fprintf(stderr, "Will kill processes using more then");
			writeMem(kill_mem, stderr);
			fprintf(stderr, " of memory.\n");
		}
		else if (strcmp(argv[loop], "-kl") == 0)
		{
			do_kill_cpu = 1;
			kill_cpu = atof(argv[loop+1]);
			if (kill_cpu <= 0 || kill_cpu > 100)
			{
				fprintf(stderr, "'kill load'-range (-kl) must be 0 < x%% <= 100.\n");
				return 1;
			}
			loop++;

			fprintf(stderr, "Will kill processes using more then %f%% of cpu-time.\n", kill_cpu);
		}
		else if (strcmp(argv[loop], "-dsql") == 0)
		{
			/* do not use special time-format */
			free(dt_format);
			dt_format = NULL;

			/* log memory usage */
			logmem = 1;

			/* log processor usage */
			logprocess = 1;

			/* log as SQL queries */
			logtosql = 1;
		}
		else if (strcmp(argv[loop], "-d") == 0)
		{
			char buffer[256];
			time_t dummy = time(NULL);
			struct tm *tm = localtime(&dummy);

			dt_format = strdup(argv[loop+1]);
			loop++;

			strftime(buffer, 255, dt_format, tm);
			fprintf(stderr, "Time format-string: %s, example: %s\n", dt_format, buffer);
		}
		else if (strcmp(argv[loop], "-m") == 0)
		{
			exit_load = atof(argv[loop+1]);
			if (exit_load <= 0.0)
			{
				fprintf(stderr, "exit-load must be bigger then 0. To disable this feature, do not specify any value.\n");
				return 1;
			}
			loop++;

			fprintf(stderr, "Exit-load: %f\n", exit_load);
		}
		else if (strcmp(argv[loop], "-o") == 0)
		{
			if (strcmp(argv[loop+1], "-") == 0)
			{
				filename = strdup(argv[loop+1]);
			}
			else
			{
				char buffer[MAXPATHLEN];
				FILE *fh;

				/* first, do an open & close to make sure the file exists */
				fh = fopen(argv[loop+1], "a+");
				if (!fh)
				{
					fprintf(stderr, "Problem opening file %s [%d]", argv[loop+1], errno);
					return 1;
				}
				fclose(fh);

				if (realpath(argv[loop+1], buffer) == NULL)
				{
					fprintf(stderr, "Problem getting realpath for file '%s' [%d]\n", argv[loop+1], errno);
					return 1;
				}

				filename = strdup(buffer);
			}
			loop++;

			fprintf(stderr, "Output will be written to: %s\n", filename);
		}
		else if (strcmp(argv[loop], "-u") == 0)
		{
			logmem = 1;

			fprintf(stderr, "Will log memory usage\n");
		}
		else if (strcmp(argv[loop], "-b") == 0)
		{
			logprocess = 1;

			fprintf(stderr, "Will log the process which used the biggest amount of CPU time in last interval.\n");
		}
		else if (strcmp(argv[loop], "-a") == 0)
		{
			shortmemamount = 1;

			fprintf(stderr, "Memory amounts will be abreviated\n");
		}
		else if (strcmp(argv[loop], "-i") == 0)
		{
			interval = atoi(argv[loop+1]);
			loop++;

			fprintf(stderr, "Sample interval: %d\n", interval);
		}
		else if (strcmp(argv[loop], "-l") == 0)
		{
			load_log = atof(argv[loop+1]);
			loop++;

			fprintf(stderr, "Will store samples when load reaches %f.\n", load_log);
		}
		else if (strcmp(argv[loop], "-f") == 0)
		{
			bg = 1;

			fprintf(stderr, "Will run in the background.\n");
		}
		else
		{
			fprintf(stderr, "Unknown option: %s\n", argv[loop]);
			return 1;
		}
	}

	/* verify sanity of given parameters */
	if (!filename)
	{
		fprintf(stderr, "You did not specify the file to write to.\n");
		return 1;
	}
	else if (strcasecmp(filename, "-") == 0 && bg)
	{
		fprintf(stderr, "Cannot write to console while running as backgroundproces.\n");
		return 1;
	}

	/* allocate memory for buffer */
	dataL = (double *)malloc(sizeof(double) * maxN);
	dataT = (time_t *)malloc(sizeof(time_t) * maxN);
	if (logmem)
	{
		dataM_p = (long *)malloc(sizeof(long) * maxN);	/* physical memory usage */
		dataM_s = (long *)malloc(sizeof(long) * maxN);	/* swap memory usage */
	}
	if (logprocess)
	{
		dataP = (char **)malloc(sizeof(char *) * maxN);	/* list of cpu eaters cmdline */
		dataPID = (pid_t *)malloc(sizeof(pid_t) * maxN);	/* pid */
		dataPERC = (float *)malloc(sizeof(float) * maxN);	/* percentage */
	}

	if (dataL == NULL || dataT == NULL || (logmem == 1 && (dataM_p == NULL || dataM_s == NULL)) || (logprocess == 1 && (dataP == NULL || dataPID == NULL || dataPERC == NULL)))
	{
		fprintf(stderr, "Problem allocating enough memory!\n");
		return 254;
	}

	/* run as bg-proc */
	if (bg)
	{
		int pid;

		/* fork */
		if ((pid = fork()) == -1)
		{
			syslog(LOG_ERR, "cannot fork! [%d]", errno);
			return 1;
		}

		/* parent? */
		if (pid)
		{
			fprintf(stderr, "Started successfully. Problems will be logged trough syslogd.\n");
			return 0;
		}
        
		/* de-tach from current TTY */
		if (setsid() == -1)
		{
			/* failed! */
			syslog(LOG_ALERT, "Cannot detach from TTY\n");
			return 1;
		}

		/* one had to do another fork after setting new process-group. can't remember why */
		if (fork()) return 0;

		/* change cur-dir to / to prevent directory from keeping locked ('unmount' should ring a bell) */
		if (chdir("/") == -1)
		{
			/* failed! */
			syslog(LOG_ALERT, "Cannot change dir. to '/' [%d]", errno);
			return 1;
		}

		/* we want full control over permissions of anything we write */
		(void)umask(0); /* this call allways seems to succeed */

		/* close I/O channels */
		if (close(0) == -1 || close(1) == -1 || close(2) == -1)
		{
			syslog(LOG_ALERT, "Error while closing I/O-channels! [%d]", errno);
			return 1;
		}
	}

	/* install signalhandler */
	signal(SIGHUP, sighandler);
	signal(SIGTERM, sighandler);

	/* main loop */
	for(;;)
	{
		int	sleep_time_len = interval;

		/* get sample */
		dataL[N] = loadAvg();
		dataT[N] = time(NULL);
		if (logmem)
		{
			memInUse(&dataM_p[N], &dataM_s[N]);
		}

		/* examine processes */
		if (scanProc() == -1)
		{
			syslog(LOG_ERR, "Internal error!");
			exit(254);
		}

		/* only examine percentages when intresting */
		if (logprocess || do_kill_cpu)
		{
			int index_biggest;

			if (calc_cputime_percentages(curproclist, prevproclist, &index_biggest) == -1)
			{
				/* first run */
				dataP[N] = NULL;
				dataPID[N] = -1;
				dataPERC[N] = 0.0;
			}
			else
			{
				dataP[N] = strdup((curproclist -> pl)[index_biggest].cmd);
				dataPID[N] = (curproclist -> pl)[index_biggest].pid;
				dataPERC[N] = (curproclist -> pl)[index_biggest].percentage;
			}
		}

		/* load too high? exit */
		if (dataL[N] >= exit_load && exit_load > 0.0)
		{
			syslog(LOG_ALERT, "Load too high, exit.");
			exit(255);
		}

		/* shall we remember this sample? */
		if (dataL[N] >= load_log)
		{
			N++;
		}

		/* buffer full? write to file */
		if (N ==  maxN)
		{
			writedata();
		}

		/* wait a while */
		while (sleep_time_len) sleep_time_len = sleep(sleep_time_len); /* sleep returns # seconds left */
	}

	return 0;
}
